/*
 *	emu.c
 *
 *	GameBoy/EMU
 *
 *	* Sat Feb 19 21:54:00 JST 2005 Naoyuki Sawa
 *	- 쐬JnB
 */
#include "app.h"

/****************************************************************************
 *	
 ****************************************************************************/

int
gb_init(const char* fname)
{
	stream_stop();  /* ĂȂ͂ǁAÔߊmɒ~ */

	/* MBC܂B */
	if(mbc_init(fname) != 0) {
		goto L_ERR;
	}

	/* ́A^C}NA܂B */
	INPUT = 0;
	DIV = 0;
	TIM = 0;

	/* VideoASoundZbg܂B
	 * - CPUZbgGameBoy[IPLAWX^̂ŁA
	 *   KCPUZbgɁAӑuZbgĂB
	 */
	dmgvideo_reset(dmgmem);
	dmgsound2_reset();

	/* CPUZbg܂B */
	dmgcpu_reset(gb_read, gb_write, 1/*GameBoy[IPL*/);

	return 0;

L_ERR:
	return -1;
}

void
input_update()
{
	int pad = pcePadGetDirect();
	INPUT = 0;
	if(pad & PAD_RI    ) INPUT |= (1<<0);
	if(pad & PAD_LF    ) INPUT |= (1<<1);
	if(pad & PAD_UP    ) INPUT |= (1<<2);
	if(pad & PAD_DN    ) INPUT |= (1<<3);
	if(pad & PAD_A     ) INPUT |= (1<<4);
	if(pad & PAD_B     ) INPUT |= (1<<5);
	if(pad & PAD_SELECT) INPUT |= (1<<6);
	if(pad & PAD_START ) INPUT |= (1<<7);
}

void
timer_update()
{
	int tima;

	/* JE^XVB(16384[Hz]=256[Cycle/Count]) */
	DIV += 456;
	dmgmem[0xff04] = DIV >> 8;			/* $FF04 (DIV) */

	/* 荞݃^C}XVB */
	if(dmgmem[0xff07] & (1<<2)) {			/* $FF07 (TAC) */
		TIM += 456;
		tima = dmgmem[0xff05];			/* $FF05 (TIMA) */
		switch(dmgmem[0xff07] & 3) {		/* $FF07 (TAC) */
		case 1:       tima += (TIM>> 4); TIM &= (1<< 4)-1; break;
		case 2:       tima += (TIM>> 6); TIM &= (1<< 6)-1; break;
		case 3:       tima += (TIM>> 8); TIM &= (1<< 8)-1; break;
		default/*0*/: tima += (TIM>>10); TIM &= (1<<10)-1; break;
		}
		if(tima >= 0x100) {
			/* ̃G~[^ł́A1C2ȏ̃^C}荞݂͐ł܂B
			 * ۂ̃vOłAȍȃ^C}荞݂͎gȂƎv܂B
			 */
			do {
				tima -= 0x100;
				tima += dmgmem[0xff06];	/* $FF06 (TMA) */
			} while(tima >= 0x100);
			dmgmem[0xff0f] |= DMGCPU_TIMER_OVERFLOW;
			dmgcpu_check_pending();
		}
		dmgmem[0xff05] = tima;			/* $FF05 (TIMA) */
	}
}

int
gb_stream_callback(short* wbuff, int param)
{
	dmgsound2_mix(wbuff);
	return 0;
}

int
gb_run(const char* fname, int sound_on)
{
	int cnt = 0;
	//
	int irq;
	int lines;
	unsigned char* draw;

#ifdef USE_STACK_SWITCH
	if(stack_switch(sram_stack + sizeof sram_stack, gb_init, (int)fname, 0)) {
#else /*USE_STACK_SWITCH*/
	if(gb_init(fname) != 0) {
#endif /*USE_STACK_SWITCH*/
		return -1;
	}

	if(sound_on) {
		stream_play(DMGSOUND2BUFLEN, gb_stream_callback, 0, 0);
	}

	cnt = 0;
	for(;;) {
		/* ͍XVB */
		input_update();

		/* t[XLbvB */
		draw = NULL;
		if((cnt += FPS) >= 60) {
			cnt -= 60;
			draw = vbuff;
		}

		/* 456[Cycle/Line]~(ʓ144[Line]+ʊO10[Line]) */
		irq = dmgvideo_new_frame(draw);
		if(irq) {
			dmgmem[0xff0f] |= irq;
			dmgcpu_check_pending();
		}
		lines = 154;
		do {
			//dmgcpu_run(dmgcpu.cycle + 456);
			//ۂɂ́AӑũANZX⊄荞݃ANmbWTCN̂߂ɁA
			//CPU456TCNtɗpłȂ͂łB
			//sTCN⏭Ȃ߂ɂāAG~[Vx҂Ƃɂ܂B
			//ɖ肪oAsTCN𒲐ĂB
			dmgcpu_run(dmgcpu.cycle + (456 - 16));

			irq = dmgvideo_next_line();
			if(irq) {
				dmgmem[0xff0f] |= irq; /* Vertical Blank & LCDC Status Interrupt */
				dmgcpu_check_pending();
			}

			/* JE^A^C}XVB */
			timer_update();
		} while(--lines);

		/* ʍXVB */
		if(draw) {
			dmgvideo_reduce(vbuff, vbuff);
			schedule();
			/* START+SELECTŁAQ[Ivsꂽ? */
			if(!mode) {
				break;
			}
		}
	}

	if(sound_on) {
		stream_stop();
	}

#ifdef USE_STACK_SWITCH
	stack_switch(sram_stack + sizeof sram_stack, mbc_exit, 0, 0);
#else /*USE_STACK_SWITCH*/
	mbc_exit();
#endif /*USE_STACK_SWITCH*/

	return 0;
}

/****************************************************************************
 *	
 ****************************************************************************/

int
gb_read_io(int addr)
{
	int data;
	switch(addr) {
	/*----- Joypad -----*/
	//case 0xff00: /* $FF00 (P1) */
	/*----- Serial -----*/
	//case 0xff01: /* $FF01 (SB) */
	//case 0xff02: /* $FF02 (SC) */
	/*----- Timer -----*/
	//case 0xff04: /* $FF04 (DIV) */
	//case 0xff05: /* $FF05 (TIMA) */
	//case 0xff06: /* $FF06 (TMA) */
	//case 0xff07: /* $FF07 (TAC) */
	/*----- Interrupt -----*/
	//case 0xff0f: /* $FF0F (IF) */
	//case 0xffff: /* $FFFF (IE) */
	/*----- Sound -----*/
	case 0xff10: /* $FF10 (NR10) */
	case 0xff11: /* $FF11 (NR11) */
	case 0xff12: /* $FF12 (NR12) */
	case 0xff13: /* $FF13 (NR13) */
	case 0xff14: /* $FF14 (NR14) */
	case 0xff16: /* $FF16 (NR21) */
	case 0xff17: /* $FF17 (NR22) */
	case 0xff18: /* $FF18 (NR23) */
	case 0xff19: /* $FF19 (NR24) */
	case 0xff1a: /* $FF1A (NR30) */
	case 0xff1b: /* $FF1B (NR31) */
	case 0xff1c: /* $FF1C (NR32) */
	case 0xff1d: /* $FF1D (NR33) */
	case 0xff1e: /* $FF1E (NR34) */
	case 0xff20: /* $FF20 (NR41) */
	case 0xff21: /* $FF21 (NR42) */
	case 0xff22: /* $FF22 (NR43) */
	case 0xff23: /* $FF23 (NR44) */
	case 0xff24: /* $FF24 (NR50) */
	case 0xff25: /* $FF25 (NR51) */
	case 0xff26: /* $FF26 (NR52) */
	case 0xff30: /* $FF30-$FF3F (Wave Pattern RAM) */
	case 0xff31: /* $FF30-$FF3F (Wave Pattern RAM) */
	case 0xff32: /* $FF30-$FF3F (Wave Pattern RAM) */
	case 0xff33: /* $FF30-$FF3F (Wave Pattern RAM) */
	case 0xff34: /* $FF30-$FF3F (Wave Pattern RAM) */
	case 0xff35: /* $FF30-$FF3F (Wave Pattern RAM) */
	case 0xff36: /* $FF30-$FF3F (Wave Pattern RAM) */
	case 0xff37: /* $FF30-$FF3F (Wave Pattern RAM) */
	case 0xff38: /* $FF30-$FF3F (Wave Pattern RAM) */
	case 0xff39: /* $FF30-$FF3F (Wave Pattern RAM) */
	case 0xff3a: /* $FF30-$FF3F (Wave Pattern RAM) */
	case 0xff3b: /* $FF30-$FF3F (Wave Pattern RAM) */
	case 0xff3c: /* $FF30-$FF3F (Wave Pattern RAM) */
	case 0xff3d: /* $FF30-$FF3F (Wave Pattern RAM) */
	case 0xff3e: /* $FF30-$FF3F (Wave Pattern RAM) */
	case 0xff3f: /* $FF30-$FF3F (Wave Pattern RAM) */
		{
			ENTER_CS;
			data = dmgsound2_read(addr);
			LEAVE_CS;
			return data;
		}
	/*----- LCDC -----*/
	case 0xff40: /* $FF40 (LCDC) */
	case 0xff41: /* $FF41 (STAT) */
	case 0xff42: /* $FF42 (SCY) */
	case 0xff43: /* $FF43 (SCX) */
	case 0xff44: /* $FF44 (LY) */
	case 0xff45: /* $FF45 (LYC) */
	case 0xff46: /* $FF46 (DMA) */
	case 0xff47: /* $FF47 (BGP) */
	case 0xff48: /* $FF48 (OBP0) */
	case 0xff49: /* $FF49 (OBP1) */
	case 0xff4a: /* $FF4A (WY) */
	case 0xff4b: /* $FF4B (WX) */
		{
			return dmgvideo_read(addr);
		}
	}
	return dmgmem[addr];
}

void
gb_write_io(int addr, int data)
{
	switch(addr) {
	/*----- Joypad -----*/
	case 0xff00: /* $FF00 (P1) */
		{
			/* P1ǂݏoD7:3$FxŒłȂ΂Ȃ悤łB
			 * (ȂƁA|Pł܂B)
			 */
			if(!(data & (1<<4))) {
				data = ~(INPUT & 15);
			} else if(!(data & (1<<5))) {
				data = ~(INPUT >> 4);
			} else {
				data = ~0; /* $Fx=GameBoy, $3x=SuperGameBoy */
			}
			dmgmem[addr] = data;
			return;
		}
	/*----- Serial -----*/
	//case 0xff01: /* $FF01 (SB) */
	case 0xff02: /* $FF02 (SC) */
		{
			if((data & (1<<7)) &&	/* Start transfer */
			   (data & (1<<0))) {	/* Internal Clock (=}X^) */
				data &= ~(1<<7);	/* M */
				dmgmem[0xff01] = 0xff;	/* Mf[^ */
				dmgmem[0xff0f] |= DMGCPU_SERIAL_TRANSFER_COMPLETION;
				dmgcpu_check_pending();
			}
			dmgmem[addr] = data;
			return;
		}
	/*----- Timer -----*/
	case 0xff04: /* $FF04 (DIV) */
		{
			dmgmem[addr] = DIV = 0; /* JE^Zbg(dl) */
			return;
		}
	//case 0xff05: /* $FF05 (TIMA) */
	//case 0xff06: /* $FF06 (TMA) */
	case 0xff07: /* $FF07 (TAC) */
		{
			if(dmgmem[addr] != data) {
				dmgmem[addr] = data;
				TIM = 0; /* [JE^Zbg */
			}
			return;
		}
	/*----- Interrupt -----*/
	case 0xff0f: /* $FF0F (IF) */
	case 0xffff: /* $FFFF (IE) */
		{
			if(dmgmem[addr] != data) {
				dmgmem[addr] = data;
				dmgcpu_check_pending();
			}
			return;
		}
	/*----- Sound -----*/
	case 0xff10: /* $FF10 (NR10) */
	case 0xff11: /* $FF11 (NR11) */
	case 0xff12: /* $FF12 (NR12) */
	case 0xff13: /* $FF13 (NR13) */
	case 0xff14: /* $FF14 (NR14) */
	case 0xff16: /* $FF16 (NR21) */
	case 0xff17: /* $FF17 (NR22) */
	case 0xff18: /* $FF18 (NR23) */
	case 0xff19: /* $FF19 (NR24) */
	case 0xff1a: /* $FF1A (NR30) */
	case 0xff1b: /* $FF1B (NR31) */
	case 0xff1c: /* $FF1C (NR32) */
	case 0xff1d: /* $FF1D (NR33) */
	case 0xff1e: /* $FF1E (NR34) */
	case 0xff20: /* $FF20 (NR41) */
	case 0xff21: /* $FF21 (NR42) */
	case 0xff22: /* $FF22 (NR43) */
	case 0xff23: /* $FF23 (NR44) */
	case 0xff24: /* $FF24 (NR50) */
	case 0xff25: /* $FF25 (NR51) */
	case 0xff26: /* $FF26 (NR52) */
	case 0xff30: /* $FF30-$FF3F (Wave Pattern RAM) */
	case 0xff31: /* $FF30-$FF3F (Wave Pattern RAM) */
	case 0xff32: /* $FF30-$FF3F (Wave Pattern RAM) */
	case 0xff33: /* $FF30-$FF3F (Wave Pattern RAM) */
	case 0xff34: /* $FF30-$FF3F (Wave Pattern RAM) */
	case 0xff35: /* $FF30-$FF3F (Wave Pattern RAM) */
	case 0xff36: /* $FF30-$FF3F (Wave Pattern RAM) */
	case 0xff37: /* $FF30-$FF3F (Wave Pattern RAM) */
	case 0xff38: /* $FF30-$FF3F (Wave Pattern RAM) */
	case 0xff39: /* $FF30-$FF3F (Wave Pattern RAM) */
	case 0xff3a: /* $FF30-$FF3F (Wave Pattern RAM) */
	case 0xff3b: /* $FF30-$FF3F (Wave Pattern RAM) */
	case 0xff3c: /* $FF30-$FF3F (Wave Pattern RAM) */
	case 0xff3d: /* $FF30-$FF3F (Wave Pattern RAM) */
	case 0xff3e: /* $FF30-$FF3F (Wave Pattern RAM) */
	case 0xff3f: /* $FF30-$FF3F (Wave Pattern RAM) */
		{
			ENTER_CS;
			dmgsound2_write(addr, data);
			LEAVE_CS;
			return;
		}
	/*----- LCDC -----*/
	case 0xff40: /* $FF40 (LCDC) */
	case 0xff41: /* $FF41 (STAT) */
	case 0xff42: /* $FF42 (SCY) */
	case 0xff43: /* $FF43 (SCX) */
	case 0xff44: /* $FF44 (LY) */
	case 0xff45: /* $FF45 (LYC) */
	case 0xff46: /* $FF46 (DMA) */
	case 0xff47: /* $FF47 (BGP) */
	case 0xff48: /* $FF48 (OBP0) */
	case 0xff49: /* $FF49 (OBP1) */
	case 0xff4a: /* $FF4A (WY) */
	case 0xff4b: /* $FF4B (WX) */
		{
			dmgvideo_write(addr, data);
			return;
		}
	}
	dmgmem[addr] = data;
}

void
gb_write_rom_bank(int addr, int data)
{
	mbc_write(addr, data);
}

